@@ -26,10 +26,28 @@
 
 static void parse_event_notifications_list(t_configuration_options *options, const char *arg);
 static void tablespace_list_append(t_configuration_options *options, const char *arg);
+static void exit_with_errors(ErrorList *config_errors);
 
+const static char *_progname = '\0';
+static char config_file_path[MAXPGPATH];
+static bool config_file_provided = false;
+static bool config_file_found = false;
+
+
+void
+set_progname(const char *argv0)
+{
+	_progname = get_progname(argv0);
+}
+
+const char *
+progname(void)
+{
+	return _progname;
+}
 
 /*
- * parse_config()
+ * load_config()
  *
  * Set default options and overwrite with values from provided configuration
  * file.
@@ -38,83 +56,161 @@ static void tablespace_list_append(t_configuration_options *options, const char
  *
  * Any configuration options changed in this function must also be changed in
  * reload_config()
+ *
+ * NOTE: this function is called before the logger is set up, so we need
+ * to handle the verbose option ourselves; also the default log level is NOTICE,
+ * so we can't use DEBUG.
  */
 bool
-parse_config(const char *config_file, t_configuration_options *options)
+load_config(const char *config_file, bool verbose, t_configuration_options *options, char *argv0)
 {
-	char	   *s,
-				buff[MAXLINELENGTH];
-	char		config_file_buf[MAXLEN];
-	char		name[MAXLEN];
-	char		value[MAXLEN];
-	bool 		config_file_provided = false;
-	FILE	   *fp;
-
-	/* Sanity checks */
+	struct stat stat_config;
 
 	/*
 	 * If a configuration file was provided, check it exists, otherwise
-	 * emit an error
+	 * emit an error and terminate. We assume that if a user explicitly
+	 * provides a configuration file, they'll want to make sure it's
+	 * used and not fall back to any of the defaults.
 	 */
 	if (config_file[0])
 	{
-		struct stat config;
+		strncpy(config_file_path, config_file, MAXPGPATH);
+		canonicalize_path(config_file_path);
 
-		strncpy(config_file_buf, config_file, MAXLEN);
-		canonicalize_path(config_file_buf);
-
-		if(stat(config_file_buf, &config) != 0)
+		if (stat(config_file_path, &stat_config) != 0)
 		{
-			log_err(_("provided configuration file '%s' not found: %s\n"),
+			log_err(_("provided configuration file \"%s\" not found: %s\n"),
 					config_file,
 					strerror(errno)
 				);
 			exit(ERR_BAD_CONFIG);
 		}
 
-		config_file_provided = true;
-	}
-
-	/*
-	 * If no configuration file was provided, set to a default file
-	 * which `parse_config()` will attempt to read if it exists
-	 */
-	else
-	{
-		strncpy(config_file_buf, DEFAULT_CONFIG_FILE, MAXLEN);
-	}
-
-
-	fp = fopen(config_file_buf, "r");
-
-	/*
-	 * Since some commands don't require a config file at all, not having one
-	 * isn't necessarily a problem.
-	 *
-	 * If the user explictly provided a configuration file and we can't
-	 * read it we'll raise an error.
-	 *
-	 * If no configuration file was provided, we'll try and read the default\
-	 * file if it exists and is readable, but won't worry if it's not.
-	 */
-	if (fp == NULL)
-	{
-		if(config_file_provided)
+		if (verbose == true)
 		{
-			log_err(_("unable to open provided configuration file '%s'; terminating\n"), config_file_buf);
-			exit(ERR_BAD_CONFIG);
+			log_notice(_("using configuration file \"%s\"\n"), config_file);
 		}
 
-		log_notice(_("no configuration file provided and default file '%s' not found - "
-					 "continuing with default values\n"),
-				   DEFAULT_CONFIG_FILE);
-		return false;
+		config_file_provided = true;
+		config_file_found = true;
 	}
 
-	/* Initialize configuration options with sensible defaults */
+	/*
+	 * If no configuration file was provided, attempt to find a default file
+	 * in this order:
+	 *  - current directory
+	 *  - /etc/repmgr.conf
+	 *  - default sysconfdir
+	 *
+	 * here we just check for the existence of the file; parse_config()
+	 * will handle read errors etc.
+	 */
+	if (config_file_provided == false)
+	{
+		char		my_exec_path[MAXPGPATH];
+		char		sysconf_etc_path[MAXPGPATH];
+
+		/* 1. "./repmgr.conf" */
+		if (verbose == true)
+		{
+			log_notice(_("looking for configuration file in current directory\n"));
+		}
+
+		snprintf(config_file_path, MAXPGPATH, "./%s", CONFIG_FILE_NAME);
+		canonicalize_path(config_file_path);
+
+		if (stat(config_file_path, &stat_config) == 0)
+		{
+			config_file_found = true;
+			goto end_search;
+		}
+
+		/* 2. "/etc/repmgr.conf" */
+		if (verbose == true)
+		{
+			log_notice(_("looking for configuration file in /etc\n"));
+		}
+
+		snprintf(config_file_path, MAXPGPATH, "/etc/%s", CONFIG_FILE_NAME);
+		if (stat(config_file_path, &stat_config) == 0)
+		{
+			config_file_found = true;
+			goto end_search;
+		}
+
+		/* 3. default sysconfdir */
+		if (find_my_exec(argv0, my_exec_path) < 0)
+		{
+			fprintf(stderr, _("%s: could not find own program executable\n"), argv0);
+			exit(EXIT_FAILURE);
+		}
+
+		get_etc_path(my_exec_path, sysconf_etc_path);
+
+		if (verbose == true)
+		{
+			log_notice(_("looking for configuration file in %s"), sysconf_etc_path);
+		}
+
+		snprintf(config_file_path, MAXPGPATH, "%s/%s", sysconf_etc_path, CONFIG_FILE_NAME);
+		if (stat(config_file_path, &stat_config) == 0)
+		{
+			config_file_found = true;
+			goto end_search;
+		}
+
+	end_search:
+		if (config_file_found == true)
+		{
+			if (verbose == true)
+			{
+				log_notice(_("configuration file found at: %s\n"), config_file_path);
+			}
+		}
+		else
+		{
+			if (verbose == true)
+			{
+				log_notice(_("no configuration file provided or found\n"));
+			}
+		}
+	}
+
+	return parse_config(options);
+}
+
+
+/*
+ * Parse configuration file; if any errors are encountered,
+ * list them and exit.
+ *
+ * Ensure any default values set here are synced with repmgr.conf.sample
+ * and any other documentation.
+ */
+bool
+parse_config(t_configuration_options *options)
+{
+	FILE	   *fp;
+	char	   *s,
+				buf[MAXLINELENGTH];
+	char		name[MAXLEN];
+	char		value[MAXLEN];
+
+	/* For sanity-checking provided conninfo string */
+	PQconninfoOption *conninfo_options;
+	char       *conninfo_errmsg = NULL;
+
+	/* Collate configuration file errors here for friendlier reporting */
+	static ErrorList config_errors = { NULL, NULL };
+
+	/* Initialize configuration options with sensible defaults
+	 * note: the default log level is set in log.c and does not need
+	 * to be initialised here
+	 */
 	memset(options->cluster_name, 0, sizeof(options->cluster_name));
 	options->node = -1;
 	options->upstream_node = NO_UPSTREAM_NODE;
+	options->use_replication_slots = 0;
 	memset(options->conninfo, 0, sizeof(options->conninfo));
 	options->failover = MANUAL_FAILOVER;
 	options->priority = DEFAULT_PRIORITY;
@@ -132,7 +228,7 @@ parse_config(const char *config_file, t_configuration_options *options)
 
 	/* default to 6 reconnection attempts at intervals of 10 seconds */
 	options->reconnect_attempts = 6;
-	options->reconnect_intvl = 10;
+	options->reconnect_interval = 10;
 
 	options->monitor_interval_secs = 2;
 	options->retry_promote_interval_secs = 300;
@@ -142,27 +238,61 @@ parse_config(const char *config_file, t_configuration_options *options)
 	options->tablespace_mapping.head = NULL;
 	options->tablespace_mapping.tail = NULL;
 
+	/*
+	 * If no configuration file available (user didn't specify and none found
+	 * in the default locations), return with default values
+	 */
+	if (config_file_found == false)
+	{
+		log_notice(_("no configuration file provided and no default file found - "
+					 "continuing with default values\n"));
+		return true;
+	}
 
+	fp = fopen(config_file_path, "r");
 
-	/* Read next line */
-	while ((s = fgets(buff, sizeof buff, fp)) != NULL)
+	/*
+	 * A configuration file has been found, either provided by the user
+	 * or found in one of the default locations. If we can't open it,
+	 * fail with an error.
+	 */
+	if (fp == NULL)
+	{
+		if (config_file_provided)
+		{
+			log_err(_("unable to open provided configuration file \"%s\"; terminating\n"), config_file_path);
+		}
+		else
+		{
+			log_err(_("unable to open default configuration file  \"%s\"; terminating\n"), config_file_path);
+		}
+
+		exit(ERR_BAD_CONFIG);
+	}
+
+	/* Read file */
+	while ((s = fgets(buf, sizeof buf, fp)) != NULL)
 	{
 		bool known_parameter = true;
 
-		/* Skip blank lines and comments */
-		if (buff[0] == '\n' || buff[0] == '#')
+		/* Parse name/value pair from line */
+		parse_line(buf, name, value);
+
+		/* Skip blank lines */
+		if (!strlen(name))
 			continue;
 
-		/* Parse name/value pair from line */
-		parse_line(buff, name, value);
+		/* Skip comments */
+		if (name[0] == '#')
+			continue;
 
 		/* Copy into correct entry in parameters struct */
 		if (strcmp(name, "cluster") == 0)
 			strncpy(options->cluster_name, value, MAXLEN);
 		else if (strcmp(name, "node") == 0)
-			options->node = atoi(value);
+			options->node = repmgr_atoi(value, "node", &config_errors);
 		else if (strcmp(name, "upstream_node") == 0)
-			options->upstream_node = atoi(value);
+			options->upstream_node = repmgr_atoi(value, "upstream_node", &config_errors);
 		else if (strcmp(name, "conninfo") == 0)
 			strncpy(options->conninfo, value, MAXLEN);
 		else if (strcmp(name, "rsync_options") == 0)
@@ -189,12 +319,11 @@ parse_config(const char *config_file, t_configuration_options *options)
 			}
 			else
 			{
-				log_err(_("value for 'failover' must be 'automatic' or 'manual'\n"));
-				exit(ERR_BAD_CONFIG);
+				error_list_append(&config_errors,_("value for 'failover' must be 'automatic' or 'manual'\n"));
 			}
 		}
 		else if (strcmp(name, "priority") == 0)
-			options->priority = atoi(value);
+			options->priority = repmgr_atoi(value, "priority", &config_errors);
 		else if (strcmp(name, "node_name") == 0)
 			strncpy(options->node_name, value, MAXLEN);
 		else if (strcmp(name, "promote_command") == 0)
@@ -202,11 +331,16 @@ parse_config(const char *config_file, t_configuration_options *options)
 		else if (strcmp(name, "follow_command") == 0)
 			strncpy(options->follow_command, value, MAXLEN);
 		else if (strcmp(name, "master_response_timeout") == 0)
-			options->master_response_timeout = atoi(value);
+			options->master_response_timeout = repmgr_atoi(value, "master_response_timeout", &config_errors);
+		/* 'primary_response_timeout' as synonym for 'master_response_timeout' -
+		 * we'll switch terminology in a future release (3.1?)
+		 */
+		else if (strcmp(name, "primary_response_timeout") == 0)
+			options->master_response_timeout = repmgr_atoi(value, "primary_response_timeout", &config_errors);
 		else if (strcmp(name, "reconnect_attempts") == 0)
-			options->reconnect_attempts = atoi(value);
+			options->reconnect_attempts = repmgr_atoi(value, "reconnect_attempts", &config_errors);
 		else if (strcmp(name, "reconnect_interval") == 0)
-			options->reconnect_intvl = atoi(value);
+			options->reconnect_interval = repmgr_atoi(value, "reconnect_interval", &config_errors);
 		else if (strcmp(name, "pg_bindir") == 0)
 			strncpy(options->pg_bindir, value, MAXLEN);
 		else if (strcmp(name, "pg_ctl_options") == 0)
@@ -216,11 +350,12 @@ parse_config(const char *config_file, t_configuration_options *options)
 		else if (strcmp(name, "logfile") == 0)
 			strncpy(options->logfile, value, MAXLEN);
 		else if (strcmp(name, "monitor_interval_secs") == 0)
-			options->monitor_interval_secs = atoi(value);
+			options->monitor_interval_secs = repmgr_atoi(value, "monitor_interval_secs", &config_errors);
 		else if (strcmp(name, "retry_promote_interval_secs") == 0)
-			options->retry_promote_interval_secs = atoi(value);
+			options->retry_promote_interval_secs = repmgr_atoi(value, "retry_promote_interval_secs", &config_errors);
 		else if (strcmp(name, "use_replication_slots") == 0)
-			options->use_replication_slots = atoi(value);
+			/* XXX we should have a dedicated boolean argument format */
+			options->use_replication_slots = repmgr_atoi(value, "use_replication_slots", &config_errors);
 		else if (strcmp(name, "event_notification_command") == 0)
 			strncpy(options->event_notification_command, value, MAXLEN);
 		else if (strcmp(name, "event_notifications") == 0)
@@ -239,9 +374,14 @@ parse_config(const char *config_file, t_configuration_options *options)
 		 * we want to accept those, we'd need to add stricter default checking,
 		 * as currently e.g. an empty `node` value will be converted to '0'.
 		 */
-		if(known_parameter == true && !strlen(value)) {
-			log_err(_("no value provided for parameter '%s'\n"), name);
-			exit(ERR_BAD_CONFIG);
+		if (known_parameter == true && !strlen(value)) {
+			char	   error_message_buf[MAXLEN] = "";
+			snprintf(error_message_buf,
+					 MAXLEN,
+					 _("no value provided for parameter \"%s\""),
+					 name);
+
+			error_list_append(&config_errors, error_message_buf);
 		}
 	}
 
@@ -252,45 +392,49 @@ parse_config(const char *config_file, t_configuration_options *options)
 	/* The following checks are for the presence of the parameter */
 	if (*options->cluster_name == '\0')
 	{
-		log_err(_("required parameter 'cluster' was not found\n"));
-		exit(ERR_BAD_CONFIG);
+		error_list_append(&config_errors, _("\"cluster\": parameter was not found\n"));
 	}
 
 	if (options->node == -1)
 	{
-		log_err(_("required parameter 'node' was not found\n"));
-		exit(ERR_BAD_CONFIG);
+		error_list_append(&config_errors, _("\"node\": parameter was not found\n"));
 	}
 
 	if (*options->node_name == '\0')
 	{
-		log_err(_("required parameter 'node_name' was not found\n"));
-		exit(ERR_BAD_CONFIG);
+		error_list_append(&config_errors, _("\"node_name\": parameter was not found\n"));
 	}
 
 	if (*options->conninfo == '\0')
 	{
-		log_err(_("required parameter 'conninfo' was not found\n"));
-		exit(ERR_BAD_CONFIG);
+		error_list_append(&config_errors, _("\"conninfo\": parameter was not found\n"));
+	}
+	else
+	{
+
+		/* Sanity check the provided conninfo string
+		 *
+		 * NOTE: PQconninfoParse() verifies the string format and checks for valid options
+		 * but does not sanity check values
+		 */
+		conninfo_options = PQconninfoParse(options->conninfo, &conninfo_errmsg);
+		if (conninfo_options == NULL)
+		{
+			char	   error_message_buf[MAXLEN] = "";
+			snprintf(error_message_buf,
+					 MAXLEN,
+					 _("\"conninfo\": %s"),
+					 conninfo_errmsg);
+
+			error_list_append(&config_errors, error_message_buf);
+		}
+
+		PQconninfoFree(conninfo_options);
 	}
 
-	/* The following checks are for valid parameter values */
-	if (options->master_response_timeout <= 0)
+	if (config_errors.head != NULL)
 	{
-		log_err(_("'master_response_timeout' must be greater than zero\n"));
-		exit(ERR_BAD_CONFIG);
-	}
-
-	if (options->reconnect_attempts < 0)
-	{
-		log_err(_("'reconnect_attempts' must be zero or greater\n"));
-		exit(ERR_BAD_CONFIG);
-	}
-
-	if (options->reconnect_intvl < 0)
-	{
-		log_err(_("'reconnect_interval' must be zero or greater\n"));
-		exit(ERR_BAD_CONFIG);
+		exit_with_errors(&config_errors);
 	}
 
 	return true;
@@ -305,7 +449,7 @@ trim(char *s)
 			   *s2 = &s[strlen(s) - 1];
 
 	/* If string is empty, no action needed */
-	if(s2 < s1)
+	if (s2 < s1)
 		return s;
 
 	/* Trim and delimit right side */
@@ -325,32 +469,58 @@ trim(char *s)
 }
 
 void
-parse_line(char *buff, char *name, char *value)
+parse_line(char *buf, char *name, char *value)
 {
 	int			i = 0;
 	int			j = 0;
 
 	/*
-	 * first we find the name of the parameter
+	 * Extract parameter name, if present
 	 */
 	for (; i < MAXLEN; ++i)
 	{
-		if (buff[i] != '=')
-			name[j++] = buff[i];
-		else
+
+		if (buf[i] == '=')
 			break;
+
+		switch(buf[i])
+		{
+			/* Ignore whitespace */
+			case ' ':
+			case '\n':
+			case '\r':
+			case '\t':
+				continue;
+			default:
+				name[j++] = buf[i];
+		}
 	}
 	name[j] = '\0';
 
 	/*
-	 * Now the value
+	 * Ignore any whitespace following the '=' sign
+	 */
+	for (; i < MAXLEN; ++i)
+	{
+		if (buf[i+1] == ' ')
+			continue;
+		if (buf[i+1] == '\t')
+			continue;
+
+		break;
+	}
+
+	/*
+	 * Extract parameter value
 	 */
 	j = 0;
 	for (++i; i < MAXLEN; ++i)
-		if (buff[i] == '\'')
+		if (buf[i] == '\'')
 			continue;
-		else if (buff[i] != '\n')
-			value[j++] = buff[i];
+		else if (buf[i] == '#')
+			break;
+		else if (buf[i] != '\n')
+			value[j++] = buf[i];
 		else
 			break;
 	value[j] = '\0';
@@ -358,7 +528,7 @@ parse_line(char *buff, char *name, char *value)
 }
 
 bool
-reload_config(char *config_file, t_configuration_options * orig_options)
+reload_config(t_configuration_options *orig_options)
 {
 	PGconn	   *conn;
 	t_configuration_options new_options;
@@ -369,7 +539,7 @@ reload_config(char *config_file, t_configuration_options * orig_options)
 	 */
 	log_info(_("reloading configuration file and updating repmgr tables\n"));
 
-	parse_config(config_file, &new_options);
+	parse_config(&new_options);
 	if (new_options.node == -1)
 	{
 		log_warning(_("unable to parse new configuration, retaining current configuration\n"));
@@ -412,13 +582,13 @@ reload_config(char *config_file, t_configuration_options * orig_options)
 		return false;
 	}
 
-	if (new_options.reconnect_intvl < 0)
+	if (new_options.reconnect_interval < 0)
 	{
 		log_warning(_("new value for 'reconnect_interval' must be zero or greater\n"));
 		return false;
 	}
 
-	if(strcmp(orig_options->conninfo, new_options.conninfo) != 0)
+	if (strcmp(orig_options->conninfo, new_options.conninfo) != 0)
 	{
 		/* Test conninfo string */
 		conn = establish_db_connection(new_options.conninfo, false);
@@ -438,56 +608,56 @@ reload_config(char *config_file, t_configuration_options * orig_options)
 	 */
 
 	/* cluster_name */
-	if(strcmp(orig_options->cluster_name, new_options.cluster_name) != 0)
+	if (strcmp(orig_options->cluster_name, new_options.cluster_name) != 0)
 	{
 		strcpy(orig_options->cluster_name, new_options.cluster_name);
 		config_changed = true;
 	}
 
 	/* conninfo */
-	if(strcmp(orig_options->conninfo, new_options.conninfo) != 0)
+	if (strcmp(orig_options->conninfo, new_options.conninfo) != 0)
 	{
 		strcpy(orig_options->conninfo, new_options.conninfo);
 		config_changed = true;
 	}
 
 	/* node */
-	if(orig_options->node != new_options.node)
+	if (orig_options->node != new_options.node)
 	{
 		orig_options->node = new_options.node;
 		config_changed = true;
 	}
 
 	/* failover */
-	if(orig_options->failover != new_options.failover)
+	if (orig_options->failover != new_options.failover)
 	{
 		orig_options->failover = new_options.failover;
 		config_changed = true;
 	}
 
 	/* priority */
-	if(orig_options->priority != new_options.priority)
+	if (orig_options->priority != new_options.priority)
 	{
 		orig_options->priority = new_options.priority;
 		config_changed = true;
 	}
 
 	/* node_name */
-	if(strcmp(orig_options->node_name, new_options.node_name) != 0)
+	if (strcmp(orig_options->node_name, new_options.node_name) != 0)
 	{
 		strcpy(orig_options->node_name, new_options.node_name);
 		config_changed = true;
 	}
 
 	/* promote_command */
-	if(strcmp(orig_options->promote_command, new_options.promote_command) != 0)
+	if (strcmp(orig_options->promote_command, new_options.promote_command) != 0)
 	{
 		strcpy(orig_options->promote_command, new_options.promote_command);
 		config_changed = true;
 	}
 
 	/* follow_command */
-	if(strcmp(orig_options->follow_command, new_options.follow_command) != 0)
+	if (strcmp(orig_options->follow_command, new_options.follow_command) != 0)
 	{
 		strcpy(orig_options->follow_command, new_options.follow_command);
 		config_changed = true;
@@ -504,76 +674,76 @@ reload_config(char *config_file, t_configuration_options * orig_options)
 	 */
 
 	/* rsync_options */
-	if(strcmp(orig_options->rsync_options, new_options.rsync_options) != 0)
+	if (strcmp(orig_options->rsync_options, new_options.rsync_options) != 0)
 	{
 		strcpy(orig_options->rsync_options, new_options.rsync_options);
 		config_changed = true;
 	}
 
 	/* ssh_options */
-	if(strcmp(orig_options->ssh_options, new_options.ssh_options) != 0)
+	if (strcmp(orig_options->ssh_options, new_options.ssh_options) != 0)
 	{
 		strcpy(orig_options->ssh_options, new_options.ssh_options);
 		config_changed = true;
 	}
 
 	/* master_response_timeout */
-	if(orig_options->master_response_timeout != new_options.master_response_timeout)
+	if (orig_options->master_response_timeout != new_options.master_response_timeout)
 	{
 		orig_options->master_response_timeout = new_options.master_response_timeout;
 		config_changed = true;
 	}
 
 	/* reconnect_attempts */
-	if(orig_options->reconnect_attempts != new_options.reconnect_attempts)
+	if (orig_options->reconnect_attempts != new_options.reconnect_attempts)
 	{
 		orig_options->reconnect_attempts = new_options.reconnect_attempts;
 		config_changed = true;
 	}
 
-	/* reconnect_intvl */
-	if(orig_options->reconnect_intvl != new_options.reconnect_intvl)
+	/* reconnect_interval */
+	if (orig_options->reconnect_interval != new_options.reconnect_interval)
 	{
-		orig_options->reconnect_intvl = new_options.reconnect_intvl;
+		orig_options->reconnect_interval = new_options.reconnect_interval;
 		config_changed = true;
 	}
 
 	/* pg_ctl_options */
-	if(strcmp(orig_options->pg_ctl_options, new_options.pg_ctl_options) != 0)
+	if (strcmp(orig_options->pg_ctl_options, new_options.pg_ctl_options) != 0)
 	{
 		strcpy(orig_options->pg_ctl_options, new_options.pg_ctl_options);
 		config_changed = true;
 	}
 
 	/* pg_basebackup_options */
-	if(strcmp(orig_options->pg_basebackup_options, new_options.pg_basebackup_options) != 0)
+	if (strcmp(orig_options->pg_basebackup_options, new_options.pg_basebackup_options) != 0)
 	{
 		strcpy(orig_options->pg_basebackup_options, new_options.pg_basebackup_options);
 		config_changed = true;
 	}
 
 	/* monitor_interval_secs */
-	if(orig_options->monitor_interval_secs != new_options.monitor_interval_secs)
+	if (orig_options->monitor_interval_secs != new_options.monitor_interval_secs)
 	{
 		orig_options->monitor_interval_secs = new_options.monitor_interval_secs;
 		config_changed = true;
 	}
 
 	/* retry_promote_interval_secs */
-	if(orig_options->retry_promote_interval_secs != new_options.retry_promote_interval_secs)
+	if (orig_options->retry_promote_interval_secs != new_options.retry_promote_interval_secs)
 	{
 		orig_options->retry_promote_interval_secs = new_options.retry_promote_interval_secs;
 		config_changed = true;
 	}
 
 	/* use_replication_slots */
-	if(orig_options->use_replication_slots != new_options.use_replication_slots)
+	if (orig_options->use_replication_slots != new_options.use_replication_slots)
 	{
 		orig_options->use_replication_slots = new_options.use_replication_slots;
 		config_changed = true;
 	}
 
-	if(config_changed == true)
+	if (config_changed == true)
 	{
 		log_debug(_("reload_config(): configuration has changed\n"));
 	}
@@ -586,6 +756,96 @@ reload_config(char *config_file, t_configuration_options * orig_options)
 }
 
 
+void
+error_list_append(ErrorList *error_list, char *error_message)
+{
+	ErrorListCell *cell;
+
+	cell = (ErrorListCell *) pg_malloc0(sizeof(ErrorListCell));
+
+	if (cell == NULL)
+	{
+		log_err(_("unable to allocate memory; terminating.\n"));
+		exit(ERR_BAD_CONFIG);
+	}
+
+	cell->error_message = pg_malloc0(MAXLEN);
+	strncpy(cell->error_message, error_message, MAXLEN);
+
+	if (error_list->tail)
+	{
+		error_list->tail->next = cell;
+	}
+	else
+	{
+		error_list->head = cell;
+	}
+
+	error_list->tail = cell;
+}
+
+
+/*
+ * Convert provided string to an integer using strtol;
+ * on error, if a callback is provided, pass the error message to that,
+ * otherwise exit
+ */
+int
+repmgr_atoi(const char *value, const char *config_item, ErrorList *error_list)
+{
+	char	  *endptr;
+	long	   longval = 0;
+	char	   error_message_buf[MAXLEN] = "";
+
+	/* It's possible that some versions of strtol() don't treat an empty
+	 * string as an error.
+	 */
+
+	if (*value == '\0')
+	{
+		snprintf(error_message_buf,
+				 MAXLEN,
+				 _("no value provided for \"%s\""),
+				 config_item);
+	}
+	else
+	{
+		errno = 0;
+		longval = strtol(value, &endptr, 10);
+
+		if (value == endptr || errno)
+		{
+			snprintf(error_message_buf,
+					 MAXLEN,
+					 _("\"%s\": invalid value (provided: \"%s\")"),
+					 config_item, value);
+		}
+	}
+
+	/* Currently there are no values which could be negative */
+	if (longval < 0)
+	{
+		snprintf(error_message_buf,
+				 MAXLEN,
+				 _("\"%s\" must be zero or greater (provided: %s)"),
+				 config_item, value);
+	}
+
+	/* Error message buffer is set */
+	if (error_message_buf[0] != '\0')
+	{
+		if (error_list == NULL)
+		{
+			log_err("%s\n", error_message_buf);
+			exit(ERR_BAD_CONFIG);
+		}
+
+		error_list_append(error_list, error_message_buf);
+	}
+
+	return (int32) longval;
+}
+
 
 /*
  * Split argument into old_dir and new_dir and append to tablespace mapping
@@ -602,7 +862,7 @@ tablespace_list_append(t_configuration_options *options, const char *arg)
 	const char *arg_ptr;
 
 	cell = (TablespaceListCell *) pg_malloc0(sizeof(TablespaceListCell));
-	if(cell == NULL)
+	if (cell == NULL)
 	{
 		log_err(_("unable to allocate memory; terminating\n"));
 		exit(ERR_BAD_CONFIG);
@@ -670,7 +930,7 @@ parse_event_notifications_list(t_configuration_options *options, const char *arg
 	for (arg_ptr = arg; arg_ptr <= (arg + strlen(arg)); arg_ptr++)
 	{
 		/* ignore whitespace */
-		if(*arg_ptr == ' ' || *arg_ptr == '\t')
+		if (*arg_ptr == ' ' || *arg_ptr == '\t')
 		{
 			continue;
 		}
@@ -679,13 +939,13 @@ parse_event_notifications_list(t_configuration_options *options, const char *arg
 		 * comma (or end-of-string) should mark the end of an event type -
 		 * just as long as there was something preceding it
 		 */
-		if((*arg_ptr == ',' || *arg_ptr == '\0') && event_type_buf[0] != '\0')
+		if ((*arg_ptr == ',' || *arg_ptr == '\0') && event_type_buf[0] != '\0')
 		{
 			EventNotificationListCell *cell;
 
 			cell = (EventNotificationListCell *) pg_malloc0(sizeof(EventNotificationListCell));
 
-			if(cell == NULL)
+			if (cell == NULL)
 			{
 				log_err(_("unable to allocate memory; terminating\n"));
 				exit(ERR_BAD_CONFIG);
@@ -708,7 +968,7 @@ parse_event_notifications_list(t_configuration_options *options, const char *arg
 			dst_ptr = event_type_buf;
 		}
 		/* ignore duplicated commas */
-		else if(*arg_ptr == ',')
+		else if (*arg_ptr == ',')
 		{
 			continue;
 		}
@@ -718,3 +978,21 @@ parse_event_notifications_list(t_configuration_options *options, const char *arg
 		}
 	}
 }
+
+
+
+static void
+exit_with_errors(ErrorList *config_errors)
+{
+	ErrorListCell *cell;
+
+	log_err(_("%s: following errors were found in the configuration file.\n"), progname());
+
+	for (cell = config_errors->head; cell; cell = cell->next)
+	{
+		log_err("%s\n", cell->error_message);
+	}
+
+	exit(ERR_BAD_CONFIG);
+}
+
