@@ -41,22 +41,6 @@
 #include "access/xlogdefs.h"
 #include "pqexpbuffer.h"
 
-/*
- * Struct to store node information
- */
-typedef struct s_node_info
-{
-	int			node_id;
-    int			upstream_node_id;
-	char		conninfo_str[MAXLEN];
-	XLogRecPtr	xlog_location;
-    t_server_type type;
-	bool		is_ready;
-	bool		is_visible;
-	char		slot_name[MAXLEN];
-	bool		active;
-}	t_node_info;
-
 
 
 /* Local info */
@@ -68,9 +52,7 @@ t_configuration_options master_options;
 
 PGconn	   *master_conn = NULL;
 
-const char *progname;
-
-char	   *config_file = DEFAULT_CONFIG_FILE;
+char	   *config_file = "";
 bool		verbose = false;
 bool		monitoring_history = false;
 t_node_info node_info;
@@ -81,17 +63,15 @@ char	   *pid_file = NULL;
 
 t_configuration_options config = T_CONFIGURATION_OPTIONS_INITIALIZER;
 
-static void help(const char *progname);
+static void help(void);
 static void usage(void);
 static void check_cluster_configuration(PGconn *conn);
 static void check_node_configuration(void);
 
 static void standby_monitor(void);
 static void witness_monitor(void);
-static bool check_connection(PGconn *conn, const char *type);
-static bool set_local_node_failed(void);
-
-static bool update_node_record_set_upstream(PGconn *conn, int this_node_id, int new_upstream_node_id);
+static bool check_connection(PGconn **conn, const char *type, const char *conninfo);
+static bool set_local_node_status(void);
 
 static void update_shared_memory(char *last_wal_standby_applied);
 static void update_registration(void);
@@ -147,6 +127,8 @@ main(int argc, char **argv)
 		{"monitoring-history", no_argument, NULL, 'm'},
 		{"daemonize", no_argument, NULL, 'd'},
 		{"pid-file", required_argument, NULL, 'p'},
+		{"help", no_argument, NULL, '?'},
+		{"version", no_argument, NULL, 'V'},
 		{NULL, 0, NULL, 0}
 	};
 
@@ -158,23 +140,10 @@ main(int argc, char **argv)
 	FILE	   *fd;
 
 	int			server_version_num = 0;
-	progname = get_progname(argv[0]);
 
-	if (argc > 1)
-	{
-		if (strcmp(argv[1], "--help") == 0 || strcmp(argv[1], "-?") == 0)
-		{
-			help(progname);
-			exit(SUCCESS);
-		}
-		if (strcmp(argv[1], "--version") == 0 || strcmp(argv[1], "-V") == 0)
-		{
-			printf("%s %s (PostgreSQL %s)\n", progname, REPMGR_VERSION, PG_VERSION);
-			exit(SUCCESS);
-		}
-	}
+	set_progname(argv[0]);
 
-	while ((c = getopt_long(argc, argv, "f:v:mdp:", long_options, &optindex)) != -1)
+	while ((c = getopt_long(argc, argv, "?Vf:vmdp:", long_options, &optindex)) != -1)
 	{
 		switch (c)
 		{
@@ -193,6 +162,12 @@ main(int argc, char **argv)
 			case 'p':
 				pid_file = optarg;
 				break;
+			case '?':
+				help();
+				exit(SUCCESS);
+			case 'V':
+				printf("%s %s (PostgreSQL %s)\n", progname(), REPMGR_VERSION, PG_VERSION);
+				exit(SUCCESS);
 			default:
 				usage();
 				exit(ERR_BAD_CONFIG);
@@ -208,7 +183,7 @@ main(int argc, char **argv)
 	 * which case we'll need to refactor parse_config() not to abort,
 	 * and return the error message.
 	 */
-	parse_config(config_file, &local_options);
+	load_config(config_file, verbose, &local_options, argv[0]);
 
 	if (daemonize)
 	{
@@ -238,10 +213,9 @@ main(int argc, char **argv)
 				strerror(errno));
 	}
 
-	logger_init(&local_options, progname, local_options.loglevel,
-				local_options.logfacility);
+	logger_init(&local_options, progname());
 	if (verbose)
-		logger_min_verbose(LOG_INFO);
+		logger_set_verbose();
 
 	if (log_type == REPMGR_SYSLOG)
 	{
@@ -255,6 +229,7 @@ main(int argc, char **argv)
 	}
 
 	/* Initialise the repmgr schema name */
+	/* XXX check this handles quoting properly */
 	maxlen_snprintf(repmgr_schema, "%s%s", DEFAULT_REPMGR_SCHEMA_PREFIX,
 			 local_options.cluster_name);
 
@@ -267,12 +242,12 @@ main(int argc, char **argv)
 
 	server_version_num = get_server_version(my_local_conn, NULL);
 
-	if(server_version_num < MIN_SUPPORTED_VERSION_NUM)
+	if (server_version_num < MIN_SUPPORTED_VERSION_NUM)
 	{
 		if (server_version_num > 0)
 		{
 			log_err(_("%s requires PostgreSQL %s or later\n"),
-					progname,
+					progname(),
 					MIN_SUPPORTED_VERSION) ;
 		}
 		else
@@ -287,10 +262,10 @@ main(int argc, char **argv)
 	node_info = get_node_info(my_local_conn, local_options.cluster_name, local_options.node);
 
 	/* No node record found - exit gracefully */
-	if(node_info.node_id == NODE_NOT_FOUND)
+	if (node_info.node_id == NODE_NOT_FOUND)
 	{
 		log_err(_("No metadata record found for this node - terminating\n"));
-		log_notice(_("HINT: was this node registered with 'repmgr (master|standby) register'?\n"));
+		log_hint(_("Check that 'repmgr (master|standby) register' was executed for this node\n"));
 		terminate(ERR_BAD_CONFIG);
 	}
 
@@ -319,7 +294,7 @@ main(int argc, char **argv)
 				check_cluster_configuration(my_local_conn);
 				check_node_configuration();
 
-				if (reload_config(config_file, &local_options))
+				if (reload_config(&local_options))
 				{
 					PQfinish(my_local_conn);
 					my_local_conn = establish_db_connection(local_options.conninfo, true);
@@ -328,7 +303,7 @@ main(int argc, char **argv)
 				}
 
 				/* Log startup event */
-				if(startup_event_logged == false)
+				if (startup_event_logged == false)
 				{
 					create_event_record(master_conn,
 										&local_options,
@@ -353,7 +328,7 @@ main(int argc, char **argv)
 				 */
 				do
 				{
-					if (check_connection(master_conn, "master"))
+					if (check_connection(&master_conn, "master", NULL))
 					{
 						sleep(local_options.monitor_interval_secs);
 					}
@@ -368,10 +343,10 @@ main(int argc, char **argv)
 					if (got_SIGHUP)
 					{
 						/*
-						 * if we can reload, then could need to change
+						 * if we can reload the configuration file, then could need to change
 						 * my_local_conn
 						 */
-						if (reload_config(config_file, &local_options))
+						if (reload_config(&local_options))
 						{
 							PQfinish(my_local_conn);
 							my_local_conn = establish_db_connection(local_options.conninfo, true);
@@ -415,7 +390,7 @@ main(int argc, char **argv)
 
 					appendPQExpBuffer(&errmsg,
 									  _("unable to connect to master node '%s'"),
-									  local_options.cluster_name);
+									  master_options.node_name);
 
 					log_err("%s\n", errmsg.data);
 
@@ -432,14 +407,14 @@ main(int argc, char **argv)
 				check_cluster_configuration(my_local_conn);
 				check_node_configuration();
 
-				if (reload_config(config_file, &local_options))
+				if (reload_config(&local_options))
 				{
 					PQfinish(my_local_conn);
 					my_local_conn = establish_db_connection(local_options.conninfo, true);
 					update_registration();
 				}
 				/* Log startup event */
-				if(startup_event_logged == false)
+				if (startup_event_logged == false)
 				{
 					create_event_record(master_conn,
 										&local_options,
@@ -465,7 +440,7 @@ main(int argc, char **argv)
 
 				do
 				{
-					log_debug("standby check loop...\n");
+					log_verbose(LOG_DEBUG, "standby check loop...\n");
 
 					if (node_info.type == WITNESS)
 					{
@@ -475,6 +450,7 @@ main(int argc, char **argv)
 					{
 						standby_monitor();
 					}
+
 					sleep(local_options.monitor_interval_secs);
 
 					if (got_SIGHUP)
@@ -483,7 +459,7 @@ main(int argc, char **argv)
 						 * if we can reload, then could need to change
 						 * my_local_conn
 						 */
-						if (reload_config(config_file, &local_options))
+						if (reload_config(&local_options))
 						{
 							PQfinish(my_local_conn);
 							my_local_conn = establish_db_connection(local_options.conninfo, true);
@@ -491,7 +467,7 @@ main(int argc, char **argv)
 						}
 						got_SIGHUP = false;
 					}
-					if(failover_done)
+					if (failover_done)
 					{
 						log_debug(_("standby check loop will terminate\n"));
 					}
@@ -536,9 +512,9 @@ witness_monitor(void)
 	 * of a missing master and promotion of a standby by that standby's
 	 * repmgrd, so we'll loop for a while before giving up.
 	 */
-	connection_ok = check_connection(master_conn, "master");
+	connection_ok = check_connection(&master_conn, "master", NULL);
 
-	if(connection_ok == false)
+	if (connection_ok == false)
 	{
 		int			connection_retries;
 		log_debug(_("old master node ID: %i\n"), master_options.node);
@@ -566,10 +542,10 @@ witness_monitor(void)
 			{
 				log_warning(
 					_("unable to determine a valid master server; waiting %i seconds to retry...\n"),
-					local_options.reconnect_intvl
+					local_options.reconnect_interval
 					);
 				PQfinish(master_conn);
-				sleep(local_options.reconnect_intvl);
+				sleep(local_options.reconnect_interval);
 			}
 			else
 			{
@@ -589,7 +565,7 @@ witness_monitor(void)
 			}
 		}
 
-		if(connection_ok == false)
+		if (connection_ok == false)
 		{
 			PQExpBufferData errmsg;
 			initPQExpBuffer(&errmsg);
@@ -682,6 +658,7 @@ standby_monitor(void)
 	char		last_wal_standby_received[MAXLEN];
 	char		last_wal_standby_applied[MAXLEN];
 	char		last_wal_standby_applied_timestamp[MAXLEN];
+	bool		last_wal_standby_received_gte_replayed;
 	char		sqlquery[QUERY_STR_LEN];
 
 	XLogRecPtr	lsn_master;
@@ -693,6 +670,7 @@ standby_monitor(void)
 	bool		did_retry = false;
 
 	PGconn	   *upstream_conn;
+	char		upstream_conninfo[MAXCONNINFO];
 	int			upstream_node_id;
 	t_node_info upstream_node;
 
@@ -704,33 +682,26 @@ standby_monitor(void)
 	 * no point in doing much else anyway
 	 */
 
-	if (!check_connection(my_local_conn, "standby"))
+	if (!check_connection(&my_local_conn, "standby", NULL))
 	{
 		PQExpBufferData errmsg;
 
-		set_local_node_failed();
+		set_local_node_status();
 
 		initPQExpBuffer(&errmsg);
 
 		appendPQExpBuffer(&errmsg,
-						  _("failed to connect to local node, node marked as failed and terminating!"));
+						  _("failed to connect to local node, node marked as failed!"));
 
 		log_err("%s\n", errmsg.data);
 
-		create_event_record(master_conn,
-							&local_options,
-							local_options.node,
-							"repmgrd_shutdown",
-							false,
-							errmsg.data);
-
-		terminate(ERR_DB_CON);
+		goto continue_monitoring_standby;
 	}
 
 	upstream_conn = get_upstream_connection(my_local_conn,
 											local_options.cluster_name,
 											local_options.node,
-											&upstream_node_id, NULL);
+											&upstream_node_id, upstream_conninfo);
 
 	type = upstream_node_id == master_options.node
 		? "master"
@@ -742,11 +713,11 @@ standby_monitor(void)
 	 * we cannot reconnect, try to get a new upstream node.
 	 */
 
-	check_connection(upstream_conn, type);	/* this takes up to
-											 * local_options.reconnect_attempts
-											 * local_options.reconnect_intvl seconds
-											 */
-
+	check_connection(&upstream_conn, type, upstream_conninfo);
+	/*
+	 * This takes up to local_options.reconnect_attempts *
+	 * local_options.reconnect_interval seconds
+	 */
 
 	if (PQstatus(upstream_conn) != CONNECTION_OK)
 	{
@@ -816,7 +787,7 @@ standby_monitor(void)
 			 */
 			upstream_node = get_node_info(my_local_conn, local_options.cluster_name, node_info.upstream_node_id);
 
-            if(upstream_node.type == MASTER)
+            if (upstream_node.type == MASTER)
             {
                 log_debug(_("failure detected on master node (%i); attempting to promote a standby\n"),
                           node_info.upstream_node_id);
@@ -827,7 +798,7 @@ standby_monitor(void)
                 log_debug(_("failure detected on upstream node %i; attempting to reconnect to new upstream node\n"),
                           node_info.upstream_node_id);
 
-				if(!do_upstream_standby_failover(upstream_node))
+				if (!do_upstream_standby_failover(upstream_node))
 				{
 					PQExpBufferData errmsg;
 					initPQExpBuffer(&errmsg);
@@ -853,6 +824,7 @@ standby_monitor(void)
 
 	PQfinish(upstream_conn);
 
+  continue_monitoring_standby:
 	/* Check if we still are a standby, we could have been promoted */
 	do
 	{
@@ -868,10 +840,13 @@ standby_monitor(void)
 				 * will require manual resolution as there's no way of determing
 				 * which master is the correct one.
 				 *
+				 * We should log a message so the user knows of the situation at hand.
+				 *
 				 * XXX check if the original master is still active and display a
 				 * warning
 				 */
-				log_err(_("It seems like we have been promoted, so exit from monitoring...\n"));
+				log_err(_("It seems this server was promoted manually (not by repmgr) so you might by in the presence of a split-brain.\n"));
+				log_err(_("Check your cluster and manually fix any anomaly.\n"));
 				terminate(1);
 				break;
 
@@ -879,10 +854,13 @@ standby_monitor(void)
 				log_err(_("standby node has disappeared, trying to reconnect...\n"));
 				did_retry = true;
 
-				if (!check_connection(my_local_conn, "standby"))
+				if (!check_connection(&my_local_conn, "standby", NULL))
 				{
-					set_local_node_failed();
-					terminate(0);
+					set_local_node_status();
+					/*
+					 * Let's continue checking, and if the postgres server on the
+					 * standby comes back up, we will activate it again
+					 */
 				}
 
 				break;
@@ -891,14 +869,20 @@ standby_monitor(void)
 
 	if (did_retry)
 	{
-		log_info(_("standby connection recovered!\n"));
+		/*
+		 * There's a possible situation where the standby went down for some reason
+		 * (maintenance for example) and is now up and maybe connected once again to
+		 * the stream. If we set the local standby node as failed and it's now running
+		 * and receiving replication data, we should activate it again.
+		 */
+	        set_local_node_status();
+	        log_info(_("standby connection recovered!\n"));
 	}
 
 	/* Fast path for the case where no history is requested */
 	if (!monitoring_history)
 		return;
 
-
 	/*
 	 * If original master has gone away we'll need to get the new one
 	 * from the upstream node to write monitoring information
@@ -924,7 +908,7 @@ standby_monitor(void)
 		return;
 	}
 
-	if(PQntuples(res) == 0)
+	if (PQntuples(res) == 0)
 	{
 		log_err(_("standby_monitor(): no active master found\n"));
 		PQclear(res);
@@ -934,18 +918,19 @@ standby_monitor(void)
 	active_master_id = atoi(PQgetvalue(res, 0, 0));
 	PQclear(res);
 
-	if(active_master_id != master_options.node)
+	if (active_master_id != master_options.node)
 	{
 		log_notice(_("connecting to active master (node %i)...\n"), active_master_id); \
-		if(master_conn != NULL)
+		if (master_conn != NULL)
 		{
 			PQfinish(master_conn);
 		}
 		master_conn = get_master_connection(my_local_conn,
 											 local_options.cluster_name,
 											 &master_options.node, NULL);
-
 	}
+	if (PQstatus(master_conn) != CONNECTION_OK)
+		PQreset(master_conn);
 
 	/*
 	 * Cancel any query that is still being executed, so i can insert the
@@ -959,7 +944,8 @@ standby_monitor(void)
 	/* Get local xlog info */
 	sqlquery_snprintf(sqlquery,
 					  "SELECT CURRENT_TIMESTAMP, pg_last_xlog_receive_location(), "
-					  "pg_last_xlog_replay_location(), pg_last_xact_replay_timestamp() ");
+					  "pg_last_xlog_replay_location(), pg_last_xact_replay_timestamp(), "
+					  "pg_last_xlog_receive_location() >= pg_last_xlog_replay_location()");
 
 	res = PQexec(my_local_conn, sqlquery);
 	if (PQresultStatus(res) != PGRES_TUPLES_OK)
@@ -974,10 +960,30 @@ standby_monitor(void)
 	strncpy(last_wal_standby_received, PQgetvalue(res, 0, 1), MAXLEN);
 	strncpy(last_wal_standby_applied, PQgetvalue(res, 0, 2), MAXLEN);
 	strncpy(last_wal_standby_applied_timestamp, PQgetvalue(res, 0, 3), MAXLEN);
+	last_wal_standby_received_gte_replayed = (strcmp(PQgetvalue(res, 0, 4), "t") == 0)
+		? true
+		: false;
+
 	PQclear(res);
 
+	/*
+	 * In the unusual event of a standby becoming disconnected from the primary,
+	 * while this repmgrd remains connected to the primary,  subtracting
+	 * "lsn_standby_applied" from "lsn_standby_received" and coercing to
+	 * (long long unsigned int) will result in a meaningless, very large
+	 * value which will overflow a BIGINT column and spew error messages into the
+	 * PostgreSQL log. In the absence of a better strategy, skip attempting
+	 * to insert a monitoring record.
+	 */
+	if (last_wal_standby_received_gte_replayed == false)
+	{
+		log_verbose(LOG_WARNING,
+					"Invalid replication_lag value calculated - is this standby connected to its upstream?\n");
+		return;
+	}
+
 	/* Get master xlog info */
-	sqlquery_snprintf(sqlquery, "SELECT pg_current_xlog_location()");
+	sqlquery_snprintf(sqlquery, "SELECT pg_catalog.pg_current_xlog_location()");
 
 	res = PQexec(master_conn, sqlquery);
 	if (PQresultStatus(res) != PGRES_TUPLES_OK)
@@ -1004,13 +1010,13 @@ standby_monitor(void)
 					  "            last_monitor_time, last_apply_time, "
 					  "            last_wal_primary_location, last_wal_standby_location, "
 					  "            replication_lag, apply_lag ) "
-					  "      VALUES(%d, %d, "
-					  "             '%s'::TIMESTAMP WITH TIME ZONE, '%s'::TIMESTAMP WITH TIME ZONE, "
-					  "             '%s', '%s', "
-					  "             %llu, %llu) ",
+					  "     VALUES(%d, %d, "
+					  "            '%s'::TIMESTAMP WITH TIME ZONE, '%s'::TIMESTAMP WITH TIME ZONE, "
+					  "            '%s', '%s', "
+					  "            %llu, %llu) ",
 					  get_repmgr_schema_quoted(master_conn),
 					  master_options.node, local_options.node,
-					   monitor_standby_timestamp, last_wal_standby_applied_timestamp,
+					  monitor_standby_timestamp, last_wal_standby_applied_timestamp,
 					  last_wal_master_location, last_wal_standby_received,
 					  (long long unsigned int)(lsn_master - lsn_standby_received),
 					  (long long unsigned int)(lsn_standby_received - lsn_standby_applied));
@@ -1019,7 +1025,8 @@ standby_monitor(void)
 	 * Execute the query asynchronously, but don't check for a result. We will
 	 * check the result next time we pause for a monitor step.
 	 */
-	log_debug("standby_monitor: %s\n", sqlquery);
+	log_verbose(LOG_DEBUG, "standby_monitor:() %s\n", sqlquery);
+
 	if (PQsendQuery(master_conn, sqlquery) == 0)
 		log_warning(_("query could not be sent to master. %s\n"),
 					PQerrorMessage(master_conn));
@@ -1061,10 +1068,10 @@ do_master_failover(void)
 	t_node_info nodes[FAILOVER_NODES_MAX_CHECK];
 
     /* Store details of the failed node here */
-    t_node_info failed_master = {-1, NO_UPSTREAM_NODE, "", InvalidXLogRecPtr, UNKNOWN, false, false};
+    t_node_info failed_master = T_NODE_INFO_INITIALIZER;
 
 	/* Store details of the best candidate for promotion to master here */
-	t_node_info best_candidate = {-1, NO_UPSTREAM_NODE, "", InvalidXLogRecPtr, UNKNOWN, false, false};
+	t_node_info best_candidate = T_NODE_INFO_INITIALIZER;
 
 	/* get a list of standby nodes, including myself */
 	sprintf(sqlquery,
@@ -1108,7 +1115,7 @@ do_master_failover(void)
 
 		/* Copy details of the failed node */
 		/* XXX only node_id is actually used later */
-		if(nodes[i].type == MASTER)
+		if (nodes[i].type == MASTER)
 		{
 			failed_master.node_id = nodes[i].node_id;
 			failed_master.xlog_location = nodes[i].xlog_location;
@@ -1152,8 +1159,8 @@ do_master_failover(void)
 			  total_nodes, visible_nodes);
 
 	/*
-	 * am i on the group that should keep alive? if i see less than half of
-	 * total_nodes then i should do nothing
+	 * Am I on the group that should keep alive? If I see less than half of
+	 * total_nodes then I should do nothing
 	 */
 	if (visible_nodes < (total_nodes / 2.0))
 	{
@@ -1193,12 +1200,13 @@ do_master_failover(void)
 			terminate(ERR_FAILOVER_FAIL);
 		}
 
-		sqlquery_snprintf(sqlquery, "SELECT pg_last_xlog_receive_location()");
+		sqlquery_snprintf(sqlquery, "SELECT pg_catalog.pg_last_xlog_receive_location()");
 		res = PQexec(node_conn, sqlquery);
 		if (PQresultStatus(res) != PGRES_TUPLES_OK)
 		{
 			log_info(_("unable to retrieve node's last standby location: %s\n"),
 					 PQerrorMessage(node_conn));
+
 			log_debug(_("connection details: %s\n"), nodes[i].conninfo_str);
 			PQclear(res);
 			PQfinish(node_conn);
@@ -1214,7 +1222,7 @@ do_master_failover(void)
 
 		/* If position is 0/0, error */
 		/* XXX do we need to terminate ourselves if the queried node has a problem? */
-		if(xlog_recptr == InvalidXLogRecPtr)
+		if (xlog_recptr == InvalidXLogRecPtr)
 		{
 			log_err(_("InvalidXLogRecPtr detected on standby node %i\n"), nodes[i].node_id);
 			terminate(ERR_FAILOVER_FAIL);
@@ -1224,7 +1232,7 @@ do_master_failover(void)
 	}
 
 	/* last we get info about this node, and update shared memory */
-	sprintf(sqlquery, "SELECT pg_last_xlog_receive_location()");
+	sprintf(sqlquery, "SELECT pg_catalog.pg_last_xlog_receive_location()");
 	res = PQexec(my_local_conn, sqlquery);
 	if (PQresultStatus(res) != PGRES_TUPLES_OK)
 	{
@@ -1290,7 +1298,7 @@ do_master_failover(void)
 			res = PQexec(node_conn, sqlquery);
 			if (PQresultStatus(res) != PGRES_TUPLES_OK)
 			{
-				log_err(_("PQexec failed: %s.\nReport an invalid value to not"
+				log_err(_("PQexec failed: %s.\nReport an invalid value to not "
 						  "be considered as new master and exit.\n"),
 						PQerrorMessage(node_conn));
 				PQclear(res);
@@ -1304,12 +1312,12 @@ do_master_failover(void)
 			 * empty string; otherwise position is 0/0 and we need to continue
 			 * looping until a valid LSN is reported
 			 */
-			if(xlog_recptr == InvalidXLogRecPtr)
+			if (xlog_recptr == InvalidXLogRecPtr)
 			{
-				if(lsn_format_ok == false)
+				if (lsn_format_ok == false)
 				{
 					/* Unable to parse value returned by `repmgr_get_last_standby_location()` */
-					if(*PQgetvalue(res, 0, 0) == '\0')
+					if (*PQgetvalue(res, 0, 0) == '\0')
 					{
 						log_crit(
 							_("unable to obtain LSN from node %i"), nodes[i].node_id
@@ -1342,6 +1350,9 @@ do_master_failover(void)
 				PQclear(res);
 
 				/* If position is 0/0, keep checking */
+				/* XXX we should add a timeout here to prevent infinite looping
+				 * if the other node's repmgrd is not up
+				 */
 				continue;
 			}
 
@@ -1419,8 +1430,7 @@ do_master_failover(void)
 		/* wait */
 		sleep(5);
 
-		if (verbose)
-			log_info(_("this node is the best candidate to be the new master, promoting...\n"));
+		log_notice(_("this node is the best candidate to be the new master, promoting...\n"));
 
 		log_debug(_("promote command is: \"%s\"\n"),
 				  local_options.promote_command);
@@ -1469,10 +1479,8 @@ do_master_failover(void)
 		/* wait */
 		sleep(10);
 
-		if (verbose)
-			log_info(_("node %d is the best candidate to be the new master, we should follow it...\n"),
-					 best_candidate.node_id);
-		log_debug(_("follow command is: \"%s\"\n"), local_options.follow_command);
+		log_info(_("node %d is the best candidate for new master, attempting to follow...\n"),
+				 best_candidate.node_id);
 
 		/*
 		 * The new master may some time to be promoted. The follow command
@@ -1483,57 +1491,23 @@ do_master_failover(void)
 			fflush(stderr);
 		}
 
-		/*
-		 * If 9.4 or later, and replication slots in use, we'll need to create a
-		 * slot on the new master
-		 */
-		new_master_conn = establish_db_connection(best_candidate.conninfo_str, true);
 
-		if(local_options.use_replication_slots)
-		{
-			if(create_replication_slot(new_master_conn, node_info.slot_name) == false)
-			{
-
-				appendPQExpBuffer(&event_details,
-								  _("Unable to create slot '%s' on the master node: %s"),
-								  node_info.slot_name,
-								  PQerrorMessage(new_master_conn));
-
-				log_err("%s\n", event_details.data);
-
-				create_event_record(new_master_conn,
-									&local_options,
-									node_info.node_id,
-									"repmgrd_failover_follow",
-									false,
-									event_details.data);
-
-				PQfinish(new_master_conn);
-				terminate(ERR_DB_QUERY);
-			}
-		}
+		log_debug(_("executing follow command: \"%s\"\n"), local_options.follow_command);
 
 		r = system(local_options.follow_command);
 		if (r != 0)
-		{
-			log_err(_("follow command failed. You could check and try it manually.\n"));
-			terminate(ERR_BAD_CONFIG);
-		}
-
-		/* and reconnect to the local database */
-		my_local_conn = establish_db_connection(local_options.conninfo, true);
-
-		/* update node information to reflect new status */
-		if(update_node_record_set_upstream(new_master_conn, node_info.node_id, best_candidate.node_id) == false)
 		{
 			appendPQExpBuffer(&event_details,
-							  _("Unable to update node record for node %i (following new upstream node %i)"),
-							  node_info.node_id,
-							  best_candidate.node_id);
+							  _("Unable to execute follow command:\n %s"),
+							  local_options.follow_command);
 
 			log_err("%s\n", event_details.data);
 
-			create_event_record(new_master_conn,
+			/* It won't be possible to write to the event notification
+			 * table but we should be able to generate an external notification
+			 * if required.
+			 */
+			create_event_record(NULL,
 								&local_options,
 								node_info.node_id,
 								"repmgrd_failover_follow",
@@ -1543,13 +1517,20 @@ do_master_failover(void)
 			terminate(ERR_BAD_CONFIG);
 		}
 
+		/* and reconnect to the local database */
+		my_local_conn = establish_db_connection(local_options.conninfo, true);
+
 		/* update internal record for this node*/
+		new_master_conn = establish_db_connection(best_candidate.conninfo_str, true);
+
 		node_info = get_node_info(new_master_conn, local_options.cluster_name, local_options.node);
 		appendPQExpBuffer(&event_details,
 						  _("Node %i now following new upstream node %i"),
 						  node_info.node_id,
 						  best_candidate.node_id);
 
+		log_info("%s\n", event_details.data);
+
 		create_event_record(new_master_conn,
 							&local_options,
 							node_info.node_id,
@@ -1576,6 +1557,8 @@ do_master_failover(void)
  * It might be worth providing a selection of reconnection strategies
  * as different behaviour might be desirable in different situations;
  * or maybe the option not to reconnect might be required?
+ *
+ * XXX check this handles replication slots gracefully
  */
 static bool
 do_upstream_standby_failover(t_node_info upstream_node)
@@ -1584,6 +1567,7 @@ do_upstream_standby_failover(t_node_info upstream_node)
 	char		sqlquery[QUERY_STR_LEN];
 	int			upstream_node_id = node_info.upstream_node_id;
 	int			r;
+	PQExpBufferData event_details;
 
 	log_debug(_("do_upstream_standby_failover(): performing failover for node %i\n"),
               node_info.node_id);
@@ -1592,7 +1576,7 @@ do_upstream_standby_failover(t_node_info upstream_node)
 	 * Verify that we can still talk to the cluster master even though
 	 * node upstream is not available
 	 */
-	if (!check_connection(master_conn, "master"))
+	if (!check_connection(&master_conn, "master", NULL))
 	{
 		log_err(_("do_upstream_standby_failover(): Unable to connect to last known master node\n"));
 		return false;
@@ -1616,7 +1600,7 @@ do_upstream_standby_failover(t_node_info upstream_node)
 			return false;
 		}
 
-		if(PQntuples(res) == 0)
+		if (PQntuples(res) == 0)
 		{
 			log_err(_("no node with id %i found"), upstream_node_id);
 			PQclear(res);
@@ -1624,7 +1608,7 @@ do_upstream_standby_failover(t_node_info upstream_node)
 		}
 
 		/* upstream node is inactive */
-		if(strcmp(PQgetvalue(res, 0, 1), "f") == 0)
+		if (strcmp(PQgetvalue(res, 0, 1), "f") == 0)
 		{
 			/*
 			 * Upstream node is an inactive master, meaning no there are no direct
@@ -1634,7 +1618,7 @@ do_upstream_standby_failover(t_node_info upstream_node)
 			 * provide an option to either try and find the current master and/or
 			 * a strategy to connect to a different upstream node
 			 */
-			if(strcmp(PQgetvalue(res, 0, 4), "master") == 0)
+			if (strcmp(PQgetvalue(res, 0, 4), "master") == 0)
 			{
 				log_err(_("unable to find active master node\n"));
 				PQclear(res);
@@ -1653,26 +1637,65 @@ do_upstream_standby_failover(t_node_info upstream_node)
 		}
 
 		PQclear(res);
-		sleep(local_options.reconnect_intvl);
+		sleep(local_options.reconnect_interval);
 	}
 
 	/* Close the connection to this server */
 	PQfinish(my_local_conn);
 	my_local_conn = NULL;
 
+	initPQExpBuffer(&event_details);
+
 	/* Follow new upstream */
 	r = system(local_options.follow_command);
 	if (r != 0)
 	{
-		log_err(_("follow command failed. You could check and try it manually.\n"));
+		appendPQExpBuffer(&event_details,
+						  _("Unable to execute follow command:\n %s"),
+						  local_options.follow_command);
+
+		log_err("%s\n", event_details.data);
+
+		/* It won't be possible to write to the event notification
+		 * table but we should be able to generate an external notification
+		 * if required.
+		 */
+		create_event_record(NULL,
+							&local_options,
+							node_info.node_id,
+							"repmgrd_failover_follow",
+							false,
+							event_details.data);
 		terminate(ERR_BAD_CONFIG);
 	}
 
-	if(update_node_record_set_upstream(master_conn, node_info.node_id, upstream_node_id) == false)
+	if (update_node_record_set_upstream(master_conn, local_options.cluster_name, node_info.node_id, upstream_node_id) == false)
 	{
+		appendPQExpBuffer(&event_details,
+						  _("Unable to set node %i's new upstream ID to %i"),
+						  node_info.node_id,
+						  upstream_node_id);
+		create_event_record(NULL,
+							&local_options,
+							node_info.node_id,
+							"repmgrd_failover_follow",
+							false,
+							event_details.data);
 		terminate(ERR_BAD_CONFIG);
 	}
 
+	appendPQExpBuffer(&event_details,
+					  _("Node %i is now following upstream node %i"),
+					  node_info.node_id,
+					  upstream_node_id);
+
+	create_event_record(NULL,
+						&local_options,
+						node_info.node_id,
+						"repmgrd_failover_follow",
+						true,
+						event_details.data);
+
 	my_local_conn = establish_db_connection(local_options.conninfo, true);
 
 	return true;
@@ -1681,24 +1704,33 @@ do_upstream_standby_failover(t_node_info upstream_node)
 
 
 static bool
-check_connection(PGconn *conn, const char *type)
+check_connection(PGconn **conn, const char *type, const char *conninfo)
 {
 	int			connection_retries;
 
 	/*
 	 * Check if the node is still available if after
-	 * local_options.reconnect_attempts * local_options.reconnect_intvl
+	 * local_options.reconnect_attempts * local_options.reconnect_interval
 	 * seconds of retries we cannot reconnect return false
 	 */
 	for (connection_retries = 0; connection_retries < local_options.reconnect_attempts; connection_retries++)
 	{
-		if (!is_pgup(conn, local_options.master_response_timeout))
+		if (*conn == NULL)
+		{
+			if (conninfo == NULL)
+			{
+				log_err("INTERNAL ERROR: *conn == NULL && conninfo == NULL");
+				terminate(ERR_INTERNAL);
+			}
+			*conn = establish_db_connection(conninfo, false);
+		}
+		if (!is_pgup(*conn, local_options.master_response_timeout))
 		{
 			log_warning(_("connection to %s has been lost, trying to recover... %i seconds before failover decision\n"),
 						type,
-						(local_options.reconnect_intvl * (local_options.reconnect_attempts - connection_retries)));
-			/* wait local_options.reconnect_intvl seconds between retries */
-			sleep(local_options.reconnect_intvl);
+						(local_options.reconnect_interval * (local_options.reconnect_attempts - connection_retries)));
+			/* wait local_options.reconnect_interval seconds between retries */
+			sleep(local_options.reconnect_interval);
 		}
 		else
 		{
@@ -1710,9 +1742,9 @@ check_connection(PGconn *conn, const char *type)
 		}
 	}
 
-	if (!is_pgup(conn, local_options.master_response_timeout))
+	if (!is_pgup(*conn, local_options.master_response_timeout))
 	{
-		log_err(_("unable to reconnect to %s after %i seconds...\n"),
+		log_err(_("unable to reconnect to %s (timeout %i seconds)...\n"),
 				type,
 				local_options.master_response_timeout
 			);
@@ -1725,7 +1757,7 @@ check_connection(PGconn *conn, const char *type)
 
 
 /*
- * set_local_node_failed()
+ * set_local_node_status()
  *
  * If failure of the local node is detected, attempt to connect
  * to the current master server (as stored in the global variable
@@ -1733,16 +1765,16 @@ check_connection(PGconn *conn, const char *type)
  */
 
 static bool
-set_local_node_failed(void)
+set_local_node_status(void)
 {
-	PGresult   *res;
+        PGresult       *res;
 	char		sqlquery[QUERY_STR_LEN];
-	int			active_master_node_id = NODE_NOT_FOUND;
+	int		active_master_node_id = NODE_NOT_FOUND;
 	char		master_conninfo[MAXLEN];
 
-	if (!check_connection(master_conn, "master"))
+	if (!check_connection(&master_conn, "master", NULL))
 	{
-		log_err(_("set_local_node_failed(): Unable to connect to last known master node\n"));
+		log_err(_("set_local_node_status(): Unable to connect to last known master node\n"));
 		return false;
 	}
 
@@ -1768,7 +1800,7 @@ set_local_node_failed(void)
 		return false;
 	}
 
-	if(!PQntuples(res))
+	if (!PQntuples(res))
 	{
 		log_err(_("no active master record found\n"));
 		return false;
@@ -1778,14 +1810,14 @@ set_local_node_failed(void)
 	strncpy(master_conninfo, PQgetvalue(res, 0, 1), MAXLEN);
 	PQclear(res);
 
-	if(active_master_node_id != master_options.node)
+	if (active_master_node_id != master_options.node)
 	{
 		log_notice(_("current active master is %i; attempting to connect\n"),
 			active_master_node_id);
 		PQfinish(master_conn);
 		master_conn = establish_db_connection(master_conninfo, false);
 
-		if(PQstatus(master_conn) != CONNECTION_OK)
+		if (PQstatus(master_conn) != CONNECTION_OK)
 		{
 			log_err(_("unable to connect to active master\n"));
 			return false;
@@ -1796,17 +1828,16 @@ set_local_node_failed(void)
 
 
 	/*
-	 * Attempt to set own record as inactive
+	 * Attempt to set the active record to the correct value.
+	 * First
 	 */
-	sqlquery_snprintf(sqlquery,
-					  "UPDATE %s.repl_nodes "
-					  "   SET active = FALSE "
-					  " WHERE id = %i ",
-					  get_repmgr_schema_quoted(master_conn),
-					  node_info.node_id);
 
-	res = PQexec(master_conn, sqlquery);
-	if (PQresultStatus(res) != PGRES_COMMAND_OK)
+	if (!update_node_record_status(master_conn,
+					    local_options.cluster_name,
+					    node_info.node_id,
+					    "standby",
+					    node_info.upstream_node_id,
+					    is_standby(my_local_conn)==1))
 	{
 		log_err(_("unable to set local node %i as inactive on master: %s\n"),
 				node_info.node_id,
@@ -1831,7 +1862,7 @@ check_cluster_configuration(PGconn *conn)
 	sqlquery_snprintf(sqlquery,
 					  "SELECT oid FROM pg_class "
 					  " WHERE oid = '%s.repl_nodes'::regclass ",
-					  get_repmgr_schema());
+			                  get_repmgr_schema_quoted(master_conn));
 	res = PQexec(conn, sqlquery);
 	if (PQresultStatus(res) != PGRES_TUPLES_OK)
 	{
@@ -1943,13 +1974,13 @@ lsn_to_xlogrecptr(char *lsn, bool *format_ok)
 
 	if (sscanf(lsn, "%X/%X", &xlogid, &xrecoff) != 2)
 	{
-		if(format_ok != NULL)
+		if (format_ok != NULL)
 			*format_ok = false;
 		log_err(_("incorrect log location format: %s\n"), lsn);
 		return 0;
 	}
 
-	if(format_ok != NULL)
+	if (format_ok != NULL)
 		*format_ok = true;
 
 	return (((XLogRecPtr) xlogid * 16 * 1024 * 1024 * 255) + xrecoff);
@@ -1958,25 +1989,29 @@ lsn_to_xlogrecptr(char *lsn, bool *format_ok)
 void
 usage(void)
 {
-	log_err(_("%s: Replicator manager daemon \n"), progname);
-	log_err(_("Try \"%s --help\" for more information.\n"), progname);
+	log_err(_("%s: Replicator manager daemon \n"), progname());
+	log_err(_("Try \"%s --help\" for more information.\n"), progname());
 }
 
 
 void
-help(const char *progname)
+help(void)
 {
-	printf(_("Usage: %s [OPTIONS]\n"), progname);
-	printf(_("Replicator manager daemon for PostgreSQL.\n"));
-	printf(_("\nOptions:\n"));
-	printf(_("  --help                    show this help, then exit\n"));
-	printf(_("  --version                 output version information, then exit\n"));
+	printf(_("%s: replication management daemon for PostgreSQL\n"), progname());
+	printf(_("\n"));
+	printf(_("Usage:\n"));
+	printf(_("  %s [OPTIONS]\n"), progname());
+	printf(_("\n"));
+	printf(_("Options:\n"));
+	printf(_("  -?, --help                show this help, then exit\n"));
+	printf(_("  -V, --version             output version information, then exit\n"));
 	printf(_("  -v, --verbose             output verbose activity information\n"));
 	printf(_("  -m, --monitoring-history  track advance or lag of the replication in every standby in repl_monitor\n"));
 	printf(_("  -f, --config-file=PATH    path to the configuration file\n"));
 	printf(_("  -d, --daemonize           detach process from foreground\n"));
 	printf(_("  -p, --pid-file=PATH       write a PID file\n"));
-	printf(_("\n%s monitors a cluster of servers.\n"), progname);
+	printf(_("\n"));
+	printf(_("%s monitors a cluster of servers and optionally performs failover.\n"), progname());
 }
 
 
@@ -2014,7 +2049,7 @@ terminate(int retval)
 		unlink(pid_file);
 	}
 
-	log_info(_("%s terminating...\n"), progname);
+	log_info(_("%s terminating...\n"), progname());
 
 	exit(retval);
 }
@@ -2219,23 +2254,12 @@ check_and_create_pid_file(const char *pid_file)
 t_node_info
 get_node_info(PGconn *conn, char *cluster, int node_id)
 {
-	char		sqlquery[QUERY_STR_LEN];
 	PGresult   *res;
 
-	t_node_info node_info = { NODE_NOT_FOUND, NO_UPSTREAM_NODE, "", InvalidXLogRecPtr, UNKNOWN, false, false};
+	t_node_info node_info = T_NODE_INFO_INITIALIZER;
 
-	sprintf(sqlquery,
-			"SELECT id, upstream_node_id, conninfo, type, slot_name, active "
-			"  FROM %s.repl_nodes "
-			" WHERE cluster = '%s' "
-			"   AND id = %i",
-			get_repmgr_schema_quoted(conn),
-			local_options.cluster_name,
-			node_id);
+	res = get_node_record(conn, cluster, node_id);
 
-	log_debug("get_node_info(): %s\n", sqlquery);
-
-	res = PQexec(my_local_conn, sqlquery);
 	if (PQresultStatus(res) != PGRES_TUPLES_OK)
 	{
 		PQExpBufferData errmsg;
@@ -2284,52 +2308,18 @@ get_node_info(PGconn *conn, char *cluster, int node_id)
 static t_server_type
 parse_node_type(const char *type)
 {
-	if(strcmp(type, "master") == 0)
+	if (strcmp(type, "master") == 0)
 	{
 		return MASTER;
 	}
-	else if(strcmp(type, "standby") == 0)
+	else if (strcmp(type, "standby") == 0)
 	{
 		return STANDBY;
 	}
-	else if(strcmp(type, "witness") == 0)
+	else if (strcmp(type, "witness") == 0)
 	{
 		return WITNESS;
 	}
 
 	return UNKNOWN;
 }
-
-
-static bool
-update_node_record_set_upstream(PGconn *conn, int this_node_id, int new_upstream_node_id)
-{
-	PGresult   *res;
-	char		sqlquery[QUERY_STR_LEN];
-
-	log_debug(_("update_node_record_set_upstream(): Updating node %i's upstream node to %i\n"), this_node_id, new_upstream_node_id);
-
-	sqlquery_snprintf(sqlquery,
-					  "  UPDATE %s.repl_nodes "
-					  "     SET upstream_node_id = %i "
-					  "   WHERE cluster = '%s' "
-					  "     AND id = %i ",
-					  get_repmgr_schema_quoted(conn),
-					  new_upstream_node_id,
-					  local_options.cluster_name,
-					  this_node_id);
-	res = PQexec(conn, sqlquery);
-
-	if (PQresultStatus(res) != PGRES_COMMAND_OK)
-	{
-		log_err(_("Unable to set new upstream node id: %s\n"),
-				PQerrorMessage(conn));
-		PQclear(res);
-
-		return false;
-	}
-
-	PQclear(res);
-
-	return true;
-}
