@@ -1,7 +1,7 @@
 /*
  * strutil.c
  *
- * Copyright (C) 2ndQuadrant, 2010-2014
+ * Copyright (C) 2ndQuadrant, 2010-2015
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,29 +25,21 @@
 #include "log.h"
 #include "strutil.h"
 
-static int xvsnprintf(char *str, size_t size, const char *format, va_list ap) __attribute__ ((format (PG_PRINTF_ATTRIBUTE, 3, 0)));
-
-/* Add strnlen on platforms that don't have it, like OS X */
-#ifndef strnlen
-size_t
-strnlen(const char *s, size_t n)
-{
-	const char *end = (const char *) memchr(s, '\0', n);
-	return(end ? end - s : n);
-}
-#endif
+static int
+xvsnprintf(char *str, size_t size, const char *format, va_list ap)
+__attribute__((format(PG_PRINTF_ATTRIBUTE, 3, 0)));
 
 static int
 xvsnprintf(char *str, size_t size, const char *format, va_list ap)
 {
-	int retval;
+	int			retval;
 
 	retval = vsnprintf(str, size, format, ap);
 
-	if (retval >= (int)size)
+	if (retval >= (int) size)
 	{
 		log_err(_("Buffer of size not large enough to format entire string '%s'\n"),
-		        str);
+				str);
 		exit(ERR_STR_OVERFLOW);
 	}
 
@@ -56,10 +48,10 @@ xvsnprintf(char *str, size_t size, const char *format, va_list ap)
 
 
 int
-xsnprintf(char *str, size_t size, const char *format, ...)
+xsnprintf(char *str, size_t size, const char *format,...)
 {
-	va_list arglist;
-	int retval;
+	va_list		arglist;
+	int			retval;
 
 	va_start(arglist, format);
 	retval = xvsnprintf(str, size, format, arglist);
@@ -70,7 +62,7 @@ xsnprintf(char *str, size_t size, const char *format, ...)
 
 
 int
-sqlquery_snprintf(char *str, const char *format, ...)
+sqlquery_snprintf(char *str, const char *format,...)
 {
 	va_list		arglist;
 	int			retval;
@@ -83,7 +75,8 @@ sqlquery_snprintf(char *str, const char *format, ...)
 }
 
 
-int maxlen_snprintf(char *str, const char *format, ...)
+int
+maxlen_snprintf(char *str, const char *format,...)
 {
 	va_list		arglist;
 	int			retval;
